(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf

(* Main mode *)
type mode = Explo | Conform

let parse_mode s = match String.lowercase s with
| "explo" -> Some Explo
| "conform" -> Some Conform
| _ -> None

let pp_mode = function
  | Explo -> "explo"
  | Conform -> "conform"


(* Execution *)

type mach = Local | Distant of string | Simul of string

let parse_mach s =
  let key,v =
    try
      let i = String.index s ' ' in
      String.sub s 0 i,
      String.sub s (i+1) (String.length s - (i+1))
    with Not_found -> s,"" in
  match String.lowercase key with
  | "local" -> Some Local
  | "ssh" -> Some (Distant v)
  | "memevents" -> Some (Simul v)
  | _ -> None
        

let pp_mach = function
  | Local -> "local"
  | Distant addr -> sprintf "ssh %s" addr
  | Simul opts -> sprintf "memevents %s" opts

(* Interpretation *)

type interpretation = Single | Multi

let parse_interpretation s = match String.lowercase s with
| "single" -> Some Single
| "multi" -> Some Multi
| _ -> None

let pp_interpretation = function
  | Single -> "single"
  | Multi -> "multi"

(* Checkpoint *)

let ckpt_name = "ckpt"

type t =
  {
   arch : Misc.arch ;
   output : string ; (* Directory for all output *)
   testing : string option ;
   safe : string option ;
   mode : mode ;
   mach : mach ;
   interpretation : interpretation ;
   work_dir : string ;
   nprocs : int option ;
   diy_sz : int option ;
   litmus_opts : string ;
   run_opts : string list ;
   verbose : int ;
   interactive : bool ;
   build : string ;
   stabilise : int ;
  }
    
let  default =
  {
   arch = Misc.X86 ;
   output = "." ;
   testing = None ;
   safe = None ;
   mach = Local ;
   mode = Explo ;
   interpretation = Single ;
   work_dir = "/var/tmp" ;
   nprocs = None ;
   diy_sz = None ;
   litmus_opts = "" ;
   run_opts = ["";] ;
   verbose = 0 ;
   interactive = true ;
   build = "sh comp.sh" ;
   stabilise = 5 ;
  }

let incr_verbose opts = { opts with verbose = opts.verbose + 1 ; }

let set_interactive b opts = { opts with interactive = b ; }

let set_arch arg cfg = match Misc.parse_arch arg with
| None -> raise (Arg.Bad (sprintf "unkown architecture: %s" arg))
| Some a -> { cfg with arch =a; } 

let set_mode arg cfg = match parse_mode arg with
| None -> raise (Arg.Bad (sprintf "bad mode: %s" arg))
| Some m -> { cfg with mode = m ; }

let set_nprocs i cfg = { cfg with nprocs = Some i; }

let get_nprocs a cfg = match cfg.nprocs with
| None ->
    begin
      match a with
      | Misc.PPC -> 4
      | Misc.X86 -> 2
      | Misc.ARM -> assert false
    end
| Some i -> i

