(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code
open AutoPhase
open AutoUtils

module Answers =
  Map.Make
    (struct
      type t = int
      let compare = compare
    end)

type base_ck = int StringMap.t
type answers = string Answers.t

type ckpt = 
    { logs_ck : Strings.t ;
      base_ck : base_ck ;
      lastrun_ck : int ;
      answers_ck : answers ; }

let pp_ckpt chan c =
  fprintf chan "Last Run = %i\n" c.lastrun_ck ;
  fprintf chan  "Logs = %a\n"
    (fun chan s -> Strings.pp chan "," output_string s)
    c.logs_ck ;
  flush chan 

let empty_ckpt =
  { logs_ck = Strings.empty ; base_ck = StringMap.empty ;
    lastrun_ck = -1 ; answers_ck = Answers.empty ; }

module Make(C:AutoConf.S) = struct
  open C
  module E = A.E
  module R = A.R
  module L = A.L

(* Verbose by command line *)

  let () = if verbose > 1 then MySys.debug := true 

(*********)
(* Utils *)
(*********)

  let open_process_in com =
    if verbose > 1 then eprintf "open_process_in: %s\n" com ;
    Unix.open_process_in com

  let uniq = sprintf "%i" (Unix.getpid ())

  let mk_uniq s = sprintf "%s.%s" s uniq

  let uniq_dist =
    sprintf "%s.%s"
      (Unix.gethostname ())
      uniq

  let mk_uniq_dist s = sprintf "%s.%s" s uniq_dist

(***********)
(* Parsing *)
(***********)

  let parse_relax s = R.parse_relax (LexUtil.One s)

  let parse_relaxs s =
    let rs = LexUtil.split s in
    let rs = R.parse_relaxs rs in
    let rs = R.expand_relaxs (fun _ _ -> assert false) rs in
    R.Set.of_list rs


(*********)
(* State *)
(*********)

  module Base = AutoBase.Make(A)

  module Outs = AutoOuts.Make(C)

(* Complete state *)
  type state =
      {
       nrun : int ; (* number of steps performed *)
       (* All relaxations involved *)
       all :  R.Set.t ;
       (* Initial safe set *)
       safe0 : R.Set.t ; 
       (* Cumulativity relaxation present in all *)
       cumul_relaxs : bool ;
       (* Three significant sets *)
       cur : R.Set.t ;
       safe_r : R.Set.t ;
       relaxed_r : R.Set.t ;
       (* Internal record of test subdirectories *)
       bases : Base.t ;
       (* Checkpoint data *)    
       lastrun : int ;
       logs : Strings.t ; 
       base_ntests : base_ck ;
       answers : answers ;
       (* List of tests not run again (as edge cycles) *)
       dontrun : Strings.t ;
       (* List of ok tests (as relaxations sets *)
       ok : Outs.outs ; no : Outs.outs ;
     }

  let checkpoint st =
    let name = Filename.concat my_dir AutoOpt.ckpt_name in
    let tname = sprintf "%s.tmp" name in

    let ckpt =
      { 
        logs_ck = st.logs ;
        base_ck  = st.base_ntests ;
        lastrun_ck = st.lastrun ;
        answers_ck = st.answers ;
      } in
    Misc.output_protect_gen open_out_bin
      (fun chan ->
        Marshal.to_channel chan
          ((C.opt,ckpt) : AutoOpt.t * ckpt) [])
      tname ;
    Sys.rename tname name

      
  let pp_state chan st =
    fprintf chan "** Step %i **\n" st.nrun ;
    fprintf chan "Testing: %a\n" R.pp_set st.cur ;
    fprintf chan "Relaxed: %a\n" R.pp_set st.relaxed_r ;
    fprintf chan "Safe   : %a\n%!" R.pp_set st.safe_r ;
    if verbose > 0 then Outs.pp chan "Ok" st.ok ;
    ()

  let pp_change chan tag old_set new_set =
    let more = R.Set.diff new_set old_set
    and less = R.Set.diff old_set new_set in
    let is_more = R.Set.is_empty more
    and is_less = R.Set.is_empty less in
    if not is_more then
      fprintf chan "Added %s: %a\n%!" tag R.pp_set more ;
    if not is_less then
      fprintf chan "Removed %s: %a\n%!" tag R.pp_set less ;
    ()


  let make_key phase st =
    let k =
      { Base.K.cur = st.cur ;
        Base.K.rel = st.relaxed_r ;
        Base.K.saf = st.safe_r ; } in
    { Base.Key.phase = phase ; Base.Key.key = k ; }

  let change_base key base bases st =
    { st with bases = Base.change key base bases }

(************)
(* Call diy *)
(************)

  let remove_safe_rfi safe =
    let open R in
    let open E in
    let powd d = ERS [Po (Diff,Dir W,d)] in
    let fencedwd f d =  ERS [Fenced (f,Diff,Dir W,d)] in

    R.Set.filter
      (fun r ->
        match r with
        | ERS [Rf Int; Po (Diff,_,d)] ->
            not (Set.mem (powd d) safe)
        | ERS [Rf Int; Fenced (f,Diff,_,d)] ->
            not (Set.mem (fencedwd f d) safe)
        | _ -> true)
      safe

  let as_fence r =
    let open R in
    let open E in
    match r with
    | ERS [Fenced _ as e;]
    | ERS [Rf Code.Ext;Fenced _ as e;]
    | ERS [Fenced _ as e; Rf Code.Ext;]
    | ERS [Rf Code.Ext; Fenced _ as e; Rf Code.Ext;]
      -> Some e
    | _ -> None

  let is_fence_relax f r = match as_fence r with
  | Some g -> f = g
  | None -> false

  let rfe = parse_relax "Rfe" 
  let podrr = parse_relax "PodRR"


  let rfe_is_safe safes = R.Set.mem rfe safes

  let mk_conf cumul dir base sz maxrelax relax safe =
    let fname = Filename.concat dir (sprintf "%s.conf" base) in
    Misc.output_protect
      (fun chan ->
        fprintf chan "-arch %s\n" (Misc.pp_arch A.A.arch) ;
        fprintf chan "-name %s\n" base ;
        fprintf chan "-mode critical\n";
        fprintf chan "-cumul %b\n" cumul;
        fprintf chan "-size %i\n" sz ;
        fprintf chan "-nprocs %i\n" nprocs ;
        if maxrelax > 0 then fprintf chan "-maxrelax %i\n" maxrelax ;
        let pp_rs chan rs =
          R.Set.pp chan "," 
            (fun chan r -> fprintf chan "%s" (R.pp_relax r))
            rs in
        let pp_rs_opt opt rs =
          if not (R.Set.is_empty rs) then
            fprintf chan "%s %a\n" opt pp_rs rs in
        pp_rs_opt "-relax" relax ;
        pp_rs_opt "-safe" safe ;
        ())
      fname ;
    fname

  let diy_dir = Filename.concat

  let is_directory name =
    try Sys.is_directory name
    with Sys_error _ -> false

  let mk_base_dir name =
    if is_directory name then ()
    else MySys.mkdir name

  let diy dir base st =
    let safe = st.safe_r
    and relax = st.cur in
    let dir = diy_dir dir base in

    let safe = remove_safe_rfi safe in

    mk_base_dir dir ;

    if R.Set.is_empty relax then (fun _ -> dir,0)
    else
      let do_diy cumul maxrelax = 
        let cname =
          mk_conf cumul dir base diy_sz maxrelax relax safe in
        let com =
          sprintf "cd %s && diy -conf %s" dir (Filename.basename cname) in
        let chan = open_process_in com in
        let ntests = AutoLex.diy chan in
        ignore (Unix.close_process_in chan) ;
        ntests in
      
      let rec do_rec cumul n =
        if n > 2 then
          let open AutoOpt in
          match C.opt.interpretation with
          | Single -> 0
          | Multi ->
              if cumul then 0
              else do_rec true 1
        else
          let ntests = do_diy cumul n in
          if ntests = 0 then do_rec cumul (n+1)
          else ntests in

      fun cumul ->
        let ntests = do_rec cumul 1 in
        dir,ntests

  let safe_conf dir base safe cumul =
    mk_conf cumul dir base diy_sz 0 R.Set.empty safe

  let safe_diy dir base safe cumul brfi =
    let dir = Filename.concat dir base in
    let safe = if brfi then remove_safe_rfi safe else safe in
    mk_base_dir dir ;
    if R.Set.is_empty safe then dir,0
    else
      let cname = safe_conf dir base safe cumul in
      let com =
        sprintf "cd %s && diy -conf %s" dir (Filename.basename cname) in
      let chan = open_process_in com in
      let ntests = AutoLex.diy chan in
      ignore (Unix.close_process_in chan);
      if ntests = 0 then Warn.warn_always "no safe test was generated";
      dir,ntests

(***************)
(* Call litmus *)
(***************)

  let dump_no dir dontrun =
    let fname = Filename.concat dir "no" in
    Misc.output_protect
      (fun chan ->
        Strings.iter (fprintf chan "%s\n") dontrun)
      fname

  let litmus dir base st =
    let tar_name =
      Filename.concat Filename.temp_dir_name
        (sprintf "%s.tar" (mk_uniq base)) in
    let no = Filename.concat dir "no"
    and all = Filename.concat dir "@all" in
    let com =
      sprintf "litmus %s -no %s -speedcheck true %s -o %s"
        all no litmus_opts tar_name in
    if st.nrun > st.lastrun then begin
      dump_no dir st.dontrun ;
      MySys.exec_stdout com
    end ;
    tar_name

(*************) 
(* Run tests *)
(*************)

  let dist_cleans names =
    let com =
      sprintf "/bin/rm -rf %s"
        (String.concat " " names) in
    dist_sh com

  let dist_clean dir = dist_cleans [dir]



  let upload base dname tarname =
    let dist_tar = Filename.concat dist_dir (mk_uniq_dist base ^ ".tar") in
    Sys.set_signal Sys.sigint
      (Sys.Signal_handle  (fun _ -> dist_cleans [dist_tar; dname] ; exit 1)) ;
    dist_cp tarname dist_tar ;
    let com =
      sprintf
        "/bin/rm -rf %s && mkdir %s && cd %s && tar xmf %s && /bin/rm %s && %s" 
        dname dname dname dist_tar dist_tar build in
    try dist_sh com
    with e -> dist_cleans [dist_tar; dname] ; raise e


  let run com dname dir base =

    let alloc_log base = AutoBase.alloc_log dir base in

    let rec do_run base xs st = match xs with
    | [] -> [],base,st
    | opts::rem ->
        let fname,base = alloc_log base in
        let st =
          if Strings.mem fname st.logs then st
          else begin
            let com = sprintf "cd %s && %s %s" dname com opts in
            begin try  dist_sh_save com fname
            with e ->
              if dname <> dir then dist_clean dname ;
              raise e end ;
            let logs = Strings.add fname st.logs in
            let st = { st with logs = logs ; } in
            checkpoint st ;
            st
          end in
        let fnames,base,st = do_run base rem st in
        fname::fnames,base,st in

    do_run base run_opts


  let compile_and_run base tarname st =
    let dname = Filename.concat dist_dir (mk_uniq_dist base.AutoBase.id) in
    let dir = Filename.concat my_dir base.AutoBase.id in
    if st.nrun > st.lastrun then upload base.AutoBase.id dname tarname ;
    let r = run "sh run.sh" dname dir base st in
    if st.nrun > st.lastrun then begin
      let com = sprintf "/bin/rm -rf %s" dname in
      dist_sh com
    end ;
    Sys.set_signal Sys.sigint Sys.Signal_default ;
    r

  let simul_run base opts st =
    let dir = Filename.concat my_dir base.AutoBase.id in
    run (sprintf "memevents -auto @all %s 2>/dev/null" opts) dir dir base st

(****************)
(* Analyse logs *)
(****************)
  let tested_single o =
    R.Set.union
      (R.Set.of_list o.L.relaxs)
      (R.Set.of_list o.L.safes)

  let tested_multi st o =
    R.Set.unions
      (R.SetSet.elements
         (R.relaxs_of st.all (A.E.parse_edges o.L.cycle)))

  let tested_relaxs tested_one logs =
    R.Set.unions (List.rev_map tested_one logs)

  let read_logs show_ok cumul names st  =
    let logs = L.add_files names in
    let tested_one =
      if cumul then tested_multi st
      else tested_single in
    let present = tested_relaxs tested_one logs in
    let yes,no =
      List.partition (fun o -> o.L.validates) logs in
    let dontrun =
      Strings.union
        (Strings.of_list
           (List.map (fun o -> o.L.cycle) yes))
        st.dontrun in
    
    let ok = Outs.diff (Outs.make st.all yes) st.ok in
    let new_ok = Outs.union ok st.ok in
    let outs_no = Outs.diff (Outs.make st.all no) new_ok in
    if (show_ok || verbose > 0) && yes <> []  then begin
      let chan = if show_ok then stdout else stderr in
      Outs.pp chan "Ok just read in log" ok
    end ;
    outs_no,{ st with dontrun = dontrun ; ok = new_ok ; },present

  let get_relaxed_assuming safe ok = Outs.get_relaxed_assuming safe ok


(* Safe heuristic *)
  let (++) x f = if R.Set.is_empty x then f () else x
    
  let get_safe_in_outs os st =    
    if verbose > 0 then Outs.pp stderr "No" os ;
    let os = Outs.simplify_for_safes st.relaxed_r st.cur os in
    if verbose > 0 then Outs.pp stderr "Simplified No" os ;
    let safe =
      Outs.safe_by_inter os
        ++ (fun () -> Outs.safe_by_cardinal os) in
    if verbose > 0 then begin
      eprintf "Safes from logs: %a\n%!" R.pp_set safe
    end ;
    safe

(* False safe heuristic *)

  let reduce_safe st =
    let rec do_rec safe ok =
      let ok = Outs.unexplained safe ok in
      if Outs.is_empty ok then safe
      else begin
        if verbose > 0 then Outs.pp stderr "Unexplained" ok ;
        let false_safe = Outs.false_safes st.safe0 safe ok  in
        if verbose > 0 then eprintf "False safes %a\n%!" R.pp_set false_safe ;
        do_rec (R.Set.diff safe false_safe) ok
      end in
    let safe = st.safe_r and ok = st.ok in
    let false_safe0 = R.Set.inter safe (get_relaxed_assuming st.safe0 st.ok) in
    let safe =
      if R.Set.is_empty false_safe0 then safe
      else begin
         if verbose > 0 then eprintf "False safes (exact) %a\n%!"
             R.pp_set false_safe0 ;
        (R.Set.diff safe false_safe0)
      end in
    do_rec safe ok


(********)
(* zyva *)
(********)
      
(* Generic diy & litmus & run *)

  let call_diy_litmus_run phase st cumul brfi =
    let key = make_key phase st in
    let base,bases = Base.look key st.bases in
    let base_id = base.AutoBase.id in
    let src_dir,base =
      if
        base.AutoBase.ntests < 0
      then
        if st.nrun > st.lastrun then
          let src_dir,ntests =
            match phase with
            | One -> diy my_dir base_id st cumul
            | Two -> safe_diy my_dir base_id st.safe_r cumul brfi in
          let base = { base with AutoBase.ntests = ntests ; } in
          src_dir,base
        else 
          diy_dir my_dir base_id,          
          try 
            let ntests = StringMap.find base_id st.base_ntests in
            { base with AutoBase.ntests = ntests ; }
          with Not_found -> assert false
      else diy_dir my_dir base_id,base in
    let ntests = base.AutoBase.ntests in
    printf "Phase %s in %s (%i tests)\n%!" (AutoPhase.pp phase) base_id ntests ;
    let st = change_base key base bases st in
    let st =
      { st with
        base_ntests = StringMap.add base_id  ntests st.base_ntests; } in
(* Result at last *)
    if ntests = 0 then
      None,st
    else begin
      let open AutoOpt in
      let logs,base,st = match C.mach with
      | Simul opts ->
          simul_run base opts st
      | Local|Distant _ ->
          let tar_name = litmus src_dir base_id st in
          compile_and_run base tar_name st in
      let st = change_base key base bases st in
      Some logs,st
    end



(* Sub-Step one, find new relaxations to be relaxed or safe *)

(* Returns a pair bool X state, the boolean is false when obvious fixpoint is reached
   (ie sub-phase 1 does not produce any test *)      

  let phase_one st =
    let cumul = not st.cumul_relaxs in
    let logs,st = call_diy_litmus_run One st cumul true in
    match logs with
    | None -> st
    | Some logs ->
        let no,st,present = read_logs false cumul logs st in
        printf "Actually tested: %a\n%!" R.pp_set
          (R.Set.inter st.cur present) ;
(* This log may generate new relaxes. Do it now
   [to avoid false safes, eg [PodWR Fre] X 2 Ok and X 3 No *)
        let st =
          let new_relax = get_relaxed_assuming st.safe_r st.ok in
          pp_change stdout "relax" st.relaxed_r new_relax ;
          { st with relaxed_r = new_relax ; } in
(* Now find some safes... *)
        let seen_safe =  get_safe_in_outs no st in
        let new_safe = R.Set.union st.safe_r seen_safe in
        let new_relax = get_relaxed_assuming new_safe st.ok in
        pp_change stdout "safe" st.safe_r new_safe ;
        pp_change stdout "relax" st.relaxed_r new_relax ;
        let new_cur =
          R.Set.diff
            st.all (R.Set.union new_safe new_relax) in
        { st with 
          cur = new_cur ;
          safe_r = new_safe ;
          relaxed_r =  new_relax ; }


(* Sub-Step two, test safe relaxations indeed to be safe *)


  let phase_two st =
    let cumul =
      let open AutoOpt in
      match C.opt.interpretation with
      | Multi -> true
      | Single -> not st.cumul_relaxs in
    let logs,st  = call_diy_litmus_run Two st cumul true in
    match logs with
    | None -> st
    | Some logs ->
        let _,st,present = read_logs false cumul logs st in
        let absent = R.Set.diff st.safe_r present in
        if not (R.Set.is_empty absent) then
          printf "Not tested: %a\n%!" R.pp_set absent ;
        let new_safe  = reduce_safe st in
        let new_relax = get_relaxed_assuming new_safe st.ok in
        pp_change stdout "safe" st.safe_r new_safe ;
        pp_change stdout "relax" st.relaxed_r new_relax ;
        let new_cur =
          R.Set.diff
            st.all (R.Set.union new_safe new_relax) in
        { st with 
          cur = new_cur ;
          safe_r = new_safe ;
          relaxed_r =  new_relax ; }


  let update_nrun st =
    let run_completed = st.nrun in
    let st =
      { st with
        nrun = run_completed+1 ;
        lastrun = max run_completed st.lastrun ; } in
    checkpoint st ;
    st

(* Conformance *)
  let rec conform_safe st stop =
    printf "** Step %i **\n%!" st.nrun ;
    if stop = 0 then st.ok
    else
      let cumul = true in
      let logs,st = call_diy_litmus_run Two st cumul false in
      let st = 
        match logs with
        | None -> st
        | Some logs ->
            let _,st,present = read_logs true cumul logs st in
            let absent = R.Set.diff st.safe_r present in
            if not (R.Set.is_empty absent) then
              printf "Not tested: %a\n%!" R.pp_set absent ;
            st in
      conform_safe (update_nrun st) (stop-1)

  let conform st =
    let ok = conform_safe st opt.AutoOpt.stabilise in
    if Outs.is_empty ok  then begin
      printf "Safe set %a is conform\n" R.pp_set st.safe0
    end else begin
      printf "Safe set %a is not conform\n" R.pp_set st.safe0  ;
      Outs.pp stdout "Invalidating tests" ok
    end


(* Zyva *)
  let rec do_ask_user st =
    printf "Dear user, choose a subset of safe candidate relaxations from \n%a,\nor type 'all' for the whole set\n? %!"
      R.pp_set st.cur ;
    let line =
      try
        let a = Answers.find st.nrun st.answers in
        printf "%s\n" a ; a
      with Not_found -> read_line () in
    match line with
    | "all" -> line,st.cur
    | _ ->
        try line,R.Set.inter (parse_relaxs line) st.cur
        with Misc.Fatal msg -> eprintf "%s\n%!" msg ; do_ask_user st

  let ask_user st =
    if interactive then
      let line,safe = do_ask_user st in
      safe,{ st with answers = Answers.add st.nrun line st.answers ; }
    else begin
      printf "Assuming %a to be safe\n%!" R.pp_set st.cur ;
      st.cur,st
    end


  let rec split_relaxed_composite l safes =
    match l with
    | a::s -> 
        if R.Set.mem a safes then split_relaxed_composite s safes 
        else a::(split_relaxed_composite s safes)
    | _ -> l 

  let analyse_obs_relaxs r safes =
    let open E in
    match r with
    | R.ERS [e1;e2] when (e1 = Rf Ext || e2 = Rf Ext) -> 
        R.Set.singleton r 
    | R.ERS [e1;e2;e3] when (e1 = Rf Ext) -> 
        R.Set.singleton r
    | R.ERS l -> 
        let l = List.map (fun x -> R.ERS [x]) l in
        let l' = split_relaxed_composite l safes in
        R.Set.of_list l'
    | _ -> R.Set.singleton r

  let analyse_observed_relaxs relaxs safes =
    R.Set.fold (fun r k -> 
      let rs = analyse_obs_relaxs r safes in
      R.Set.union rs k) relaxs R.Set.empty

  let transf_edge e =
    let open E in
    match e with
    | Rf Int -> Po(Same,Dir W,Dir R)
    | Dp(sl,d) -> Po(sl,Dir R,d)
    | Ctrl(sl,d) -> Po(sl,Dir R,d)
    | _ -> e

  let reduce_composite_duo e1 e2 =
    let open E in
    let e1 = transf_edge e1 in let e2 = transf_edge e2 in
    match e1,e2 with
    | Po(sl1,d1,d), Po(sl2,d',d2) when d = d' ->
        let sl = match sl1,sl2 with
        | _,Diff -> Diff
        | Diff,_ -> Diff
        | Same,Same -> Same in
        Po(sl,d1,d2)
    | Po(sl1,d1,d),Fenced(f,sl2,d',d2) 
    | Fenced(f,sl1,d1,d),Po(sl2,d',d2) when d = d' ->
        let sl = match sl1,sl2 with
        | _,Diff -> Diff
        | Diff,_ -> Diff
        | Same,Same -> Same in
        Fenced(f,sl,d1,d2)
    | Fenced(f1,sl1,d1,d),Fenced(f2,sl2,d',d2) when (d = d' && f1 = f2) ->
        let sl = match sl1,sl2 with
        | _,Diff -> Diff
        | Diff,_ -> Diff
        | Same,Same -> Same in
(*faudrait reflechir a comment faire pour f1<>f2; genre Sync plus fort que LwSync*)
        Fenced(f1,sl,d1,d2)
    | _ -> failwith "bad composition of edge" 

  let rec red_comp_list l =
    let open E in
    let open R in
    match l with
    | [] -> failwith "empty composite"
    | [e] -> ERS [e]
    | [e1;e2] when (e1 = Rf Ext || e2 = Rf Ext) -> ERS l
    | [e1;e2;e3] when (e1 = Rf Ext) -> ERS l 
    | e1::(e2::s) ->
        let e = reduce_composite_duo e1 e2 in
        red_comp_list (e::s)

  let reduce_composite_relax r = 
    let open R in 
    match r with 
    | R.ERS l -> 
        let r = red_comp_list l in 
        R.Set.singleton r
    | _ -> R.Set.singleton r

  let analyse_observed_safes safes = 
    R.Set.fold (fun s k ->
      let ss = reduce_composite_relax s in
      R.Set.union ss k) safes R.Set.empty

  let remove_cumul_st st =
    let all = R.remove_cumul st.all in
    let safe = R.remove_cumul st.safe_r in
    let relaxed = get_relaxed_assuming st.safe_r st.ok in
    let cur = R.Set.diff all (R.Set.union safe relaxed) in
    { st with
      all = all ; cur = cur ; safe_r = safe ; relaxed_r =  relaxed ; }

  let expand_cumul_st st =
    let all = R.remove_cumul st.all in
    let safe = R.expand_cumul st.safe_r in
    let relaxed = R.Set.remove rfe (R.expand_cumul st.relaxed_r) in
    let safe = R.Set.diff safe relaxed in
    let cur = R.Set.diff all (R.Set.union safe relaxed) in
    { st with
      all = all ; cur = cur ; safe_r = safe ; relaxed_r =  relaxed ; }

  let over st =
    R.Set.iter
      (fun r ->
        let title =
          sprintf "Witness(es) for relaxed %s" (R.pp_relax r) in
        let w = Outs.get_relax_witnesses st.safe_r r st.ok in
        Outs.pp stdout title w)
      st.relaxed_r ;
    let st = 
      if rfe_is_safe st.safe_r then expand_cumul_st st 
      else st in
    let relaxed = analyse_observed_relaxs st.relaxed_r st.safe_r in
    let safe = remove_safe_rfi st.safe_r in
    fprintf stdout "Observed relaxed: %a\n" R.pp_set relaxed;
    fprintf stdout "Observed safe: %a\n" R.pp_set safe ;    
    fprintf stdout "** Now checking safe set conformance **\n%!" ;
    let st_conform =
      { st with safe_r = safe ; relaxed_r = R.Set.empty ;
        cur = R.Set.empty ; all = safe ; safe0 = safe ;
        ok = Outs.empty ;
        dontrun = Strings.empty ;
      } in    
    let ok = conform_safe st_conform opt.AutoOpt.stabilise in
    fprintf stdout "Observed relaxed: %a\n" R.pp_set relaxed;
    fprintf stdout "Observed safe: %a\n" R.pp_set safe ;
    if not (Outs.is_empty ok) then begin
      fprintf stdout "!!! The safe set failed conformance check !!!\n" ;
      Outs.pp stdout "Invalidating tests" ok
    end ;
    ()


(* Compute 'observed' current set *) 
let as_fences rs =
  R.Set.fold
    (fun r k -> match as_fence r with
    | Some e -> R.Set.add (R.ERS [e]) k
    | None -> k)
  rs R.Set.empty

let reduce_current st =
  let safe_fences = as_fences st.safe_r in
  R.Set.fold
    (fun r k -> match as_fence r with
    | Some e ->
        if R.Set.mem (R.ERS [e]) safe_fences then k
        else R.Set.add r k
    | None -> R.Set.add r k)
    st.cur R.Set.empty

let observed_current st =
  if rfe_is_safe st.safe_r then reduce_current st
  else st.cur

(* Zyva *)
  let st_same st1 st2 =
    R.Set.compare st1.safe_r st2.safe_r = 0 &&
    Outs.equal st1.ok st2.ok

  let rec explo st0 stop =
    pp_state stdout st0 ;
    let st1 = phase_one st0 in
    let st2 = phase_two st1 in
    let stop = if st_same st0 st2 then stop+1 else 0 in
    if stop < opt.AutoOpt.stabilise then
      explo (update_nrun st2) stop
    else if R.Set.is_empty (observed_current st2) then
      over (update_nrun st2)
    else
      let user_safe,st2 = ask_user st2 in
      let cur =  R.Set.diff st2.cur user_safe
      and safe =  R.Set.union st2.safe_r user_safe in
      let st2 = { st2 with cur = cur ; safe_r = safe ; } in
      if R.Set.is_empty (observed_current st2) then
        over (update_nrun st2)
      else
        explo (update_nrun st2) 0


(****************)
(* Entry points *)
(****************)

  let empty_state =
    { nrun = 0 ;
      cumul_relaxs = false ;
      lastrun = -1 ;
      all = R.Set.empty ;
      safe0 = R.Set.empty ;
      cur = R.Set.empty ; 
      safe_r = R.Set.empty ; relaxed_r = R.Set.empty ; 
      bases = Base.empty ;
      base_ntests = StringMap.empty ;
      logs = Strings.empty ;
      answers = Answers.empty ;
      dontrun = Strings.empty ;
      ok = Outs.empty;  no = Outs.empty;
    }


  let restore_explo c =
    let cur = parse_relaxs C.testing
    and safe =  parse_relaxs C.safe in
    let all = R.Set.union cur safe in
    explo
      { empty_state with
        logs = c.logs_ck ;  lastrun = c.lastrun_ck ;
        base_ntests = c.base_ck ;
        answers = c.answers_ck ;
        cur = R.Set.diff cur safe ;
        safe_r = safe ; all = all ;
        safe0 = safe ;
        cumul_relaxs = R.cumul_in all ;
      } 0

  let go_explo () = restore_explo empty_ckpt

  let restore_conform c =
    let safe = parse_relaxs C.safe in
    conform 
      { empty_state with
        logs = c.logs_ck ;  lastrun = c.lastrun_ck ;
        base_ntests = c.base_ck ;
        answers = c.answers_ck ;
        safe_r = safe ;
        safe0 = safe ;
        all = safe ;
        (* Useless, but well *)
        cumul_relaxs = R.cumul_in safe ;
      }


  let go_conform () = restore_conform empty_ckpt ;
end
