(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code
open LexUtil

module Make(C:XXXCompile.S) = struct

open C.E
open C.R


let parse_relaxs = List.map parse_relax

let parse_edges = List.map parse_edge

module Sc = struct
  let choice = Code.Sc
end

let gen_sc lr ls n =
(*uses alt.ml*)
  let lr = parse_relaxs lr
  and ls = parse_relaxs ls in
  let module M = Alt.Make(Sc)(C) in
  M.gen ~relax:lr ~safe:ls n

module Critical = struct
  let choice = Code.Critical
end

let gen_critical lr ls n =
  let lr = parse_relaxs lr
  and ls = parse_relaxs ls in
  let module M = Alt.Make(Critical)(C) in
  M.gen ~relax:lr ~safe:ls n

module Thin = struct
  let choice = Code.Thin
end

let gen_thin n =
  let lr = [ERS [Rf Int]; ERS [Rf Ext]]
  and ls = [PPO] in
  let module M = Alt.Make(Thin)(C) in
  M.gen ~relax:lr ~safe:ls n

module Uni = struct
  let choice = Code.Uni
end

let gen_uni n =
  let lr = [ERS [Rf Int]; ERS [Rf Ext]]
  and ls =
    [ERS [Ws Int]; ERS [Ws Ext]; ERS [Fr Int];
     ERS [Fr Ext]; ERS [Po (Same,Irr,Irr)]] in
  let module M = Alt.Make(Uni)(C) in
  M.gen ~relax:lr ~safe:ls n


let parse_edges_opt = function
  | None -> None
  | Some xs -> Some (parse_edges xs)

let go c n (*size*) olr ols (*relax and safe lists*) olc (*one + arg*) =
  match c with
  | Sc ->
    begin match olr,ols with
    | None,None ->
      let module M = Alt.Make(Sc)(C) in M.gen n    
    | None,Some ls -> gen_sc [] ls n
    | Some lr,None -> gen_sc lr [] n
    | Some lr,Some ls -> gen_sc lr ls n
    end
  | Critical ->
    begin match olr,ols with
    | None,None ->
      let module M = Alt.Make(Critical)(C) in M.gen n    
    | None,Some ls -> gen_critical [] ls n
    | Some lr,None -> gen_critical lr [] n
    | Some lr,Some ls -> gen_critical lr ls n
    end

  | Thin -> gen_thin n 
  | Uni -> gen_uni n

end

open Misc

let split s = match s with
  | None -> None
  | Some s ->
  let splitted = LexUtil.split s in
  Some splitted

let lc = ref []

let get_arg s =
  lc := (!lc)@[s]

let norm_cmd cmd = 
  match cmd with
  | [] -> assert false
  | _::s -> 
    let rec no_conf l = 
      match l with
      | [] -> []
      | "-conf"::_::l -> no_conf l
      | s::l -> s::(no_conf l)
    in no_conf s

let exec_conf s =
  let conf = Misc.input_protect LexConf.conf s in
  let prog = Sys.argv.(0) in
  let cmd = Array.to_list Sys.argv in
  let cmd = norm_cmd cmd in
  if !Misc.verbose > 1 then
    eprintf "EXEC: %s %s\n%!" prog (String.concat " " (conf @ cmd)) ;
  ignore (Unix.execvp prog (Array.of_list (prog::conf@cmd))) ;
  ()

let () =
  let module V = SymbConstant in
  Arg.parse (Config.speclist) get_arg Config.usage_msg;
  begin
  match !Config.conf with
  | None -> ()
  | Some s -> exec_conf s
  end;
  let relax_list = split !Config.relaxs
  and safe_list = split !Config.safes 
  and one_list = if !Config.one then Some !lc else None in 
  let f = match !Config.arch with
  | PPC ->
    let module M = Make(PPCCompile.Make(V)) in
    M.go
  | X86 ->
    let module M = Make(X86Compile.Make(V)) in
    M.go
  | ARM -> Warn.fatal "ARM not implemented yet" in
  try
    f !Config.mode !Config.size relax_list safe_list one_list ;
    exit 0
  with Misc.Fatal msg ->
    eprintf "Fatal error: %s\n" msg ;
    exit 2





