(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code

module Make(C:XXXCompile.S) : sig

(* C'est la vie *)
  type edge = C.E.edge
  type relax = C.R.relax

(* Type of internal state, left abstrtact *)
  type t

(* A generator is a function 'let gen ... kont k = ...'
   that generates sequences of edges es (and information builder mk_info)
   and calls 'kont es rs k' for each
   sequence *)

  type info = (string * string) list
  type mk_info = edge list -> info * C.R.Set.t
  type generator =
      ((edge list -> mk_info -> t -> t) -> t -> t)

  type check = edge list list -> bool

(* 'all check gen' will dumps all tests produced by gen *)
   
  val all :  ?check:check -> generator -> unit


end = struct


  let base = match !Config.name with
  | Some b -> b
  | None -> Warn.fatal "You must supply -name <base>"

  module T =  Top.Make(C)
  module Env = Map.Make(String)

  open C.E 
  open C.R
  type edge = C.E.edge
  type relax = C.R.relax

(* Environment for test names *)
(* + court +elegant *)


  let fresh_name name env =
    let n = try Env.find name env with Not_found -> 0 in
    sprintf "%s%03i" name n,
    Env.add name (n+1) env
      

(**********************************)
(* Computation of cycle signature *)
(**********************************)

(*
  A bit contrieved...
  - change list to array,
  - find starting index of minimal sequence in array,
  - build the signature starting from this index.
 *)

(* Signatures as strings, for the sake of compacity *)

  let sig_of_dir = function
    | Dir W -> 'W'
    | Dir R -> 'R'
    | Irr -> '*'

  let sig_of_ie = function
    | Int -> 'i'
    | Ext -> 'e'

  let sig_of_sd = function
    | Same -> 's'
    | Diff -> 'd'

  let sig_of_edge buff = function
    | Rf ie ->
        Buffer.add_char buff '1' ;
        Buffer.add_char buff (sig_of_ie ie)
    | Fr ie ->
        Buffer.add_char buff '2' ;
        Buffer.add_char buff (sig_of_ie ie)
    | Ws ie ->
        Buffer.add_char buff '3' ;
        Buffer.add_char buff (sig_of_ie ie)
    | Po (sd,e1,e2) ->
        Buffer.add_char buff '4' ;
        Buffer.add_char buff (sig_of_sd sd) ;
        Buffer.add_char buff (sig_of_dir e1) ;
        Buffer.add_char buff (sig_of_dir e2)
    | Fenced (fe,sd,e1,e2) ->
        Buffer.add_char buff '5' ;
        Buffer.add_char buff (C.A.sig_of_fence fe) ;
        Buffer.add_char buff (sig_of_sd sd) ;
        Buffer.add_char buff (sig_of_dir e1) ;
        Buffer.add_char buff (sig_of_dir e2)
    | Dp (sd,e) ->
        Buffer.add_char buff '6' ;
        Buffer.add_char buff (sig_of_sd sd) ;
        Buffer.add_char buff (sig_of_dir e)
    | Hat ->
        Buffer.add_char buff '7'
    | Ctrl (sd,e) ->
        Buffer.add_char buff '8' ;
        Buffer.add_char buff (sig_of_sd sd) ;
        Buffer.add_char buff (sig_of_dir e)
    | RfStar ie ->
        Buffer.add_char buff '9' ;
        Buffer.add_char buff (sig_of_ie ie)

(* Check edge signatures to be pairwise distinct *)          
  let () =
    let buff = Buffer.create 8 in

    let sig_of_edge e =
      sig_of_edge buff e ;
      let r = Buffer.contents buff in
      Buffer.clear buff ;
      r in

    let all =
      fold_edges (fun e k -> (e,sig_of_edge e)::k) [] in
    if !Misc.verbose > 2 then begin
      eprintf "Signatures:\n" ;
      iter_edges
        (fun e -> eprintf "%s -> %s\n" (pp_edge e) (sig_of_edge e)) ;
      flush stderr
    end ;
    let all =
      List.sort
        (fun (_,s1) (_,s2) -> String.compare s1 s2)
        all in
    let rec uniq = function
      | []|[_] -> ()
      | (e1,s1)::((e2,s2)::_ as rem) ->
          if s1=s2 then
            Warn.fatal
              "edges %s and %s have identical signatures, change sig_of_edge\n"
              (pp_edge e1) (pp_edge e2)
          else uniq rem in
    uniq all


  let sig_of_shift =
    let buff = Buffer.create 16 in
    fun  t k ->
      let sz = Array.length t in
      assert(sz > 0) ;
      let incr i = if i+1 >= sz then 0 else i+1 in
      let rec do_rec i =
        sig_of_edge buff t.(i) ;
        let j = incr i in
        if j=k then () else do_rec j in
      do_rec k ;
      let r = Buffer.contents buff in
      Buffer.clear buff ;
      r

  let cycle_of_shift t k =
    let sz = Array.length t in
    assert(sz > 0) ;
      let incr i = if i+1 >= sz then 0 else i+1 in
      let rec do_rec i =
        let es =
          let j = incr i in
          if j=k then [] else do_rec j in
        t.(i)::es in
      do_rec k

  let find_min_shift t =

    let sz = Array.length t in
    assert (sz > 0) ;
    let incr i = if i+1 >= sz then 0 else i+1 in

    let rec c_rec k i1 i2 =
      if k <= 0 then 0
      else
        let c = Pervasives.compare t.(i1) t.(i2) in
        if c=0 then c_rec (k-1) (incr i1) (incr i2)
        else c in

    let rec find_rec k_min k =
      if k >= sz then k_min
      else if c_rec sz k k_min < 0 then find_rec k (k+1)
      else find_rec k_min (k+1) in

    find_rec 0 1

  let comp_sig es = match es with
  | [] -> "",es
  | _::_ ->
      let t = Array.of_list es in
      let k = find_min_shift t in
      sig_of_shift t k, cycle_of_shift t k

  module SigSet = MySet.Make(String)

  let have_seen es sigs =
    let xxx,es= comp_sig es in
    if SigSet.mem xxx sigs then
      true,es,sigs
    else
      false,es,SigSet.add xxx sigs

(******************)
(* Internal state *)
(******************)


  type t =
      {
       ntests : int ;        (* number of tests outputed so far *)
       sigs : SigSet.t ;     (* Signatures of compiled tests *)
       name_env : int Env.t ;  (* State for getting fresh names *)
       relaxed : C.R.SetSet.t ;
     }

  type check = edge list list -> bool
  type info = (string * string) list
  type mk_info = edge list -> info * C.R.Set.t

  type generator =
      ((edge list -> mk_info -> t -> t) -> t -> t)

  let empty_t =
    { ntests = 0 ;
      sigs = SigSet.empty ; name_env = Env.empty;
      relaxed = C.R.SetSet.empty; }

(************************************** ****)
(* Check duplicates, compile and dump test *)
(*******************************************)


(* Adapt actual filename *)
  let tar_output_protect = match !Config.tarfile with
  | None -> Misc.output_protect
  | Some _ -> fun f name -> Misc.output_protect f (Tar.outname name)

(* Compile & dump proper *)        

  let dump_test all_chan check le mk_info c res =
    let n,ns = fresh_name base res.name_env in
    let cy = pp_edges le in
    let info,relaxed = mk_info le in
    let info = ("Cycle",cy)::info in
    let t = T.test_of_cycle n ~info:info ~check:check le c in
(* Output test proper *)
    let src = sprintf "%s.litmus" n in
    tar_output_protect
      (fun chan -> T.dump_test_channel chan t)
      src ;
(* And litmus file name in @all file *)
    fprintf all_chan "%s\n" src ;
    if !Misc.verbose > 0 then eprintf "Test: %s\n" n ;
    { res with
      name_env = ns ;
      ntests = res.ntests+1;
      relaxed = C.R.SetSet.add relaxed res.relaxed; }

(* Compose duplicate checker and dumper *)
  let check_dump =
    if !Config.cannonical_only then    
      fun all_chan check es mk_info r  ->
        let es,c = C.C.resolve_edges es in
        let seen,es,sigs = have_seen es r.sigs in
        if seen then Warn.fatal "Duplicate" ;
        C.C.finish c ;
        dump_test all_chan check es mk_info c { r with sigs = sigs; } 
    else 
      fun all_chan check es mk_info r ->
        let es,c = C.C.resolve_edges es in
        C.C.finish c ;
        dump_test all_chan check es mk_info c r

  let check_dump all_chan check es mk_info res =
    if !Misc.verbose > 0 then begin
      eprintf "------------------------------------------------------\n" ;
      eprintf "Cycle: %s\n" (pp_edges es) ;
      let info,_ = mk_info es in
      List.iter
        (fun (tag,i) -> eprintf "%s: %s\n" tag i) info
    end ;
    try
      check_dump all_chan check es mk_info res
    with Misc.Fatal msg ->
      if !Misc.verbose > 0 then begin
        eprintf "Compilation failed: %s\n" msg
      end ;
      res

(* Exported *)
  let all ?(check=(fun _ -> true)) gen =
    tar_output_protect
      (fun all_chan ->
        fprintf all_chan
          "# %s\n" (String.concat " " (Array.to_list Sys.argv)) ;
        fprintf all_chan "# Revision: %s\n" Version.rev ;
        let res =  gen (check_dump all_chan check) empty_t in
        flush stderr ;
        printf
          "Generator produced %d tests\n%!"
          res.ntests ;
        printf "Relaxations tested: %a\n"
          pp_set_set res.relaxed)
      "@all" ;
    match !Config.tarfile with
    | None -> ()
    | Some name -> Tar.tar name

end
